from dataclasses import dataclass
from typing import Dict, List, Union

import torch


def ensure_correct_dims(t, batch_size):
    n_seq = int(t.flatten().shape[0] / batch_size)
    return t.reshape((batch_size, n_seq))


@dataclass
class DataCollatorForDocumentClassificationBATCH:
    def __call__(
            self, examples: List[Union[List[int], torch.Tensor, Dict[str, torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        batch_size = len(examples)
        return {
            'input_ids': ensure_correct_dims(torch.stack([e['input_ids'] for e in examples], dim=0), batch_size),
            'attention_mask': ensure_correct_dims(torch.stack([e['attention_mask'] for e in examples], dim=0),
                                                  batch_size),
            'labels': ensure_correct_dims(torch.stack([e['labels'] for e in examples], dim=0), batch_size),
        }

@dataclass
class DataCollatorForAutoencodersBATCH:
    def __call__(
            self, examples: List[Union[List[int], torch.Tensor, Dict[str, torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        batch_size = len(examples)
        return {
            'input_ids': ensure_correct_dims(torch.stack([e['input_ids'] for e in examples], dim=0), batch_size),
            'attention_mask': ensure_correct_dims(torch.stack([e['attention_mask'] for e in examples], dim=0),
                                                  batch_size),
            'labels': ensure_correct_dims(torch.stack([e['labels'] for e in examples], dim=0), batch_size),
        }


import numpy as np
import torch
from torch.nn.functional import softmax


from transformers import GPT2LMHeadModel, GPT2Config, GPT2TokenizerFast
import logging


class FluencyScorer:
    def __init__(self, tok_dir=None, model_dir=None, base_type='gpt2', device='cpu'):
        logger = logging.Logger('Fluency Scorer')
        if tok_dir and not model_dir:
            logger.warning(
                f"Custom tokenizer specified ({tok_dir}), but a generic model is being used. Please reconsider unless you know what you're doing.")
        self.tokenizer = GPT2TokenizerFast.from_pretrained(tok_dir if tok_dir else base_type)
        if not self.tokenizer.pad_token_id:
            self.tokenizer.pad_token = self.tokenizer.eos_token
        self.model = GPT2LMHeadModel.from_pretrained(model_dir if model_dir else base_type).to(device)

    def __call__(self, sentences, stride=10):
        if not isinstance(sentences, list):
            sentences = [sentences]
        return [{'perplexity': self.perplexity(s, stride=stride), 'sequence': s} for s in sentences]

    def perplexity(self, s, stride=10):
        # Thanks huggingface! https://huggingface.co/transformers/perplexity.html
        encodings = self.tokenizer(s, return_tensors='pt')
        max_length = self.model.config.n_positions

        lls = []
        for i in range(0, encodings.input_ids.size(1), stride):
            # Use a sliding window to calculate next-token log_likelihood.
            # Smaller windows are more accurate
            begin_loc = max(i + stride - max_length, 0)
            end_loc = min(i + stride, encodings.input_ids.size(1))
            trg_len = end_loc - i  # may be different from stride on last loop
            input_ids = encodings.input_ids[:, begin_loc:end_loc].to(self.model.device)
            target_ids = input_ids.clone()
            target_ids[:, :-trg_len] = -100

            with torch.no_grad():
                outputs = self.model(input_ids, labels=target_ids)
                # outputs[0] is loss in this case
                log_likelihood = outputs[0] * trg_len

            lls.append(log_likelihood)
        if lls:
            ppl = torch.exp(torch.stack(lls).sum() / end_loc)
            return ppl.item()
        else:
            # Perplexity cannot be calculated on the empty string.
            # Set to infinity to strongly disincentivise.
            return np.inf







import pandas as pd

MESSAGE = 'Hello, world!'

TC_CLASS_MAP = {
    0: '2800 - Semiconductors/Memory, Circuits/Measuring and Testing, Optics/Photocopying, Printing/Measuring and Testing',
    1: '1600 - Biotechnology and Organic Chemistry',
    2: '2600 - Communications',
    3: '3700 - Mechanical Engineering, Manufacturing, Gaming, and Medical Devices/Processes',
    4: '1700 - Chemical and Materials Engineering',
    5: '3600 - Transportation, Construction, Electronic Commerce, Agriculture, National Security and License and Review',
    6: '2100 - Computer Architecture and Software',
    7: '2400 - Networking, Multiplexing, Cable, and Security'
}

TC_COUNT_MAP = {2800: 1043341, 3700: 584685, 2600: 525022, 3600: 504068, 1700: 466615, 1600: 322352, 2100: 317652,
                2400: 265921, 2700: 14765, 3900: 2212, 4100: 994, 2900: 730, 1800: 71, 3200: 59, 2500: 54, 3300: 48,
                1200: 47, 3400: 43, 2300: 38, 3800: 37, 3100: 34, 2000: 28, 2200: 26, 3500: 21, 1100: 20, 1900: 18,
                1500: 16, 1000: 13, 7800: 13, 7600: 12, 1300: 12, 5700: 11, 3000: 6, 7700: 4, 4200: 3, 5600: 3, 4700: 2,
                4600: 2, 5100: 1, 5400: 1, 4500: 1, 9700: 1, 6500: 1, 500: 1, 6700: 1, 8600: 1, 4000: 1, 4800: 1,
                8800: 1, 1400: 1, 400: 1, 7200: 1}

TC_TOTAL = 4049012  # sum(TC_COUNT_MAP.values())

GPT3_KEY = 'sk-yM3CMLAJYwBBYwtCHGV43AVA8C5OZSMvAxjjkl4O'

BLOCKLIST = ['fig',
             'fig.',
             'figs',
             'figs.',
             'figure',
             'example',
             'examples',
             'table',
             'tables',
             'jan.',
             'feb.',
             'mar.',
             'apr.',
             'may.',
             'jun.',
             'jul.',
             'aug.',
             'sep.',
             'sept.',
             'oct.',
             'nov.',
             'dec.',
             'jan',
             'feb',
             'mar',
             'apr',
             'may',
             'jun',
             'jul',
             'aug',
             'sep',
             'sept',
             'oct',
             'nov',
             'dec',]

STRIPLIST = ['ie.',
             'i.e.',
             'eg.',
             'e.g.',]

IDEATION_PROMPT_TEMPLATE = '''D1: Aviation
D2: Air systems
Abs: A cabin air system for aviation electronics comprises a compressor, an air conditioner, a mix manifold, and an aviation electronics module. The compressor compresses ambient air, and the air conditioner cools the compressed ambient air to produce cabin air. The mix manifold delivers some of the cabin air to the electronics module.
#########
D1: Memory management
D2: Distributed computing
Abs: Computational storage techniques for distribute computing are disclosed. The computational storage server receives input from multiple clients, which is used by the server when executing one or more computation functions. The computational storage server can aggregate multiple client inputs before applying one or more computation functions. The computational storage server sets up: a first memory area for storing input received from multiple clients; a second memory area designated for storing the computation functions to be executed by the computational storage server using the input data received from the multiple clients; a client specific memory management area for storing metadata related to computations performed by the computational storage server for specific clients; and a persistent storage area for storing checkpoints associated with aggregating computations performed by the computation functions.
#########
D1: {}
D2: {}
Abs:'''



import torch
import logging

class LengthScorer:
    def __init__(self):
        logger = logging.Logger('Fluency Scorer')
    def __call__(self, source, target, c = 1):
        return c * min(len(source), len(target)) / max(len(source), len(target))


from datetime import datetime
from typing import Union

import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import CrossEntropyLoss
from torch.utils import data
from transformers import (T5ForConditionalGeneration, T5Config, T5TokenizerFast)
from transformers import (Trainer,
                          TrainingArguments)
from transformers.modeling_outputs import (
    BaseModelOutput,
    BaseModelOutputWithPastAndCrossAttentions,
    Seq2SeqLMOutput,
    Seq2SeqModelOutput,
)

from ..collators import DataCollatorForAutoencodersBATCH


class T5EncoderAggDecoder(T5ForConditionalGeneration):
    def __init__(self, config: T5Config,
                 block_size=512,
                 encoding_vector_size=512,
                 agg=True,
                 agg_mode='linear'):
        super(T5EncoderAggDecoder, self).__init__(config)
        self.block_size = block_size

        self.agg = agg
        self.agg_mode = agg_mode
        print(f'AGG: {self.agg}, MODE: {self.agg_mode if self.agg else "N/A"}')
        if self.agg and self.agg_mode == 'linear':
            print('Using Linear Aggregation')
            self.encoding_vector_size = encoding_vector_size
            self.enc_to_vec = nn.Linear(self.block_size * self.config.d_model, encoding_vector_size)
            self.vec_to_enc_hat = nn.Linear(encoding_vector_size, self.block_size * self.config.d_model)

    def forward(
            self,
            agg=True,
            input_ids=None,
            attention_mask=None,
            decoder_input_ids=None,
            decoder_attention_mask=None,
            encoder_outputs=None,
            past_key_values=None,
            head_mask=None,
            labels=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
    ):

        if self.model_parallel:
            assert False, "This variant does not support model parallel."

        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # Encode if needed (training, first prediction pass)

        encoder_outputs, dense_vector_encoding = self.encode_w_agg(  # If encoder_outputs provided, will spit back out
            encoder_outputs=encoder_outputs,
            input_ids=input_ids,
            attention_mask=attention_mask,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict, )

        hidden_states = encoder_outputs[0]

        # Decode
        decoder_outputs = self.decode_custom(
            labels=labels,
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_attention_mask,
            past_key_values=past_key_values,
            encoder_hidden_states=hidden_states,
            encoder_attention_mask=attention_mask,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = decoder_outputs[0]

        return self.lm_head_step(sequence_output, decoder_outputs, encoder_outputs, labels=labels,
                                 return_dict=return_dict)

    def lm_head_step(self, sequence_output, decoder_outputs, encoder_outputs, labels=None, return_dict=False):
        if self.config.tie_word_embeddings:
            # Rescale output before projecting on vocab
            # See https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/transformer/transformer.py#L586
            sequence_output = sequence_output * (self.model_dim ** -0.5)

        lm_logits = self.lm_head(sequence_output)

        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss(ignore_index=-100)
            loss = loss_fct(lm_logits.view(-1, lm_logits.size(-1)), labels.view(-1))
            # TODO(thom): Add z_loss https://github.com/tensorflow/mesh/blob/fa19d69eafc9a482aff0b59ddd96b025c0cb207d/mesh_tensorflow/layers.py#L666

        if not return_dict:
            output = (lm_logits,) + decoder_outputs[1:] + encoder_outputs
            return ((loss,) + output) if loss is not None else output

        return Seq2SeqLMOutput(
            loss=loss,
            logits=lm_logits,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

    def encode_w_agg(self,
                     encoder_outputs=None,
                     input_ids=None,
                     attention_mask=None,
                     head_mask=None,
                     output_attentions=False,
                     output_hidden_states=False,
                     return_dict=True,
                     **unused_kwargs
                     ):

        if encoder_outputs is None:
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                head_mask=head_mask,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )

            if self.agg:
                agg_outputs = self.temporal_agg(encoder_outputs, attention_mask)
                encoder_outputs = agg_outputs['output_formatted_for_decoder']
                dense_vector_encoding = agg_outputs['dense_vector_encoding']

        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        return encoder_outputs, dense_vector_encoding

    def temporal_agg(self, encoder_outputs, attention_masks):
        encoder_last_hidden_states = encoder_outputs.last_hidden_state
        batch_size, seq_len, d_model = encoder_last_hidden_states.shape

        result = {'output_formatted_for_decoder': None,
                  'dense_vector_encoding': None}

        # Copy mask along d_model axis (the third one)
        masks_expanded = attention_masks[:, :, None].repeat(1, 1, d_model)

        # elementwise multiplication of enc last hidden state with mask,
        # which should remove irrelevant states (those which were masked) from the average.
        hidden_masked = encoder_last_hidden_states.mul(masks_expanded)

        if self.agg_mode == 'mean':
            # Average, dividing by total unmasked tokens (time axis, e.g. the second)
            hidden_masked_summed = hidden_masked.sum(1)
            masked_time_agg = hidden_masked_summed.div(masks_expanded.sum(1))

            masked_time_agg_broadcast = masked_time_agg[:, None, :].repeat(1, seq_len, 1)

            # Make sure to put back the time axis!
            result['output_formatted_for_decoder'] = BaseModelOutput(
                last_hidden_state=masked_time_agg_broadcast,
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

            result['dense_vector_encoding'] = masked_time_agg

        elif self.agg_mode == 'linear':
            # Stack each sequence embedding and project to an encoding vector
            stacked = hidden_masked.view(batch_size, seq_len * d_model)
            dense_vector_encoding = self.enc_to_vec(stacked)  # (batch_size, self.encoding_vector_size)
            result['dense_vector_encoding'] = dense_vector_encoding
            encoder_output_hat = self.vec_to_enc_hat(dense_vector_encoding).view(batch_size, seq_len,
                                                                                 d_model)  # (batch_size, seq_len, d_model)

            result['output_formatted_for_decoder'] = BaseModelOutput(last_hidden_state=encoder_output_hat,
                                                                     hidden_states=encoder_outputs[1] if len(
                                                                         encoder_outputs) > 1 else None,
                                                                     attentions=encoder_outputs[2] if len(
                                                                         encoder_outputs) > 2 else None, )

        return result

    def _shift_right(self, input_ids):
        # In T5 decoder_start_token_id is usually set to the pad_token_id.
        decoder_start_token_id = self.config.pad_token_id
        pad_token_id = self.config.pad_token_id

        # shift inputs to the right
        shifted_input_ids = input_ids.new_zeros(input_ids.shape)
        shifted_input_ids[..., 1:] = input_ids[..., :-1].clone()
        shifted_input_ids[..., 0] = decoder_start_token_id

        assert pad_token_id is not None, "self.model.config.pad_token_id has to be defined."
        # replace possible -100 values in labels by `pad_token_id`
        shifted_input_ids.masked_fill_(shifted_input_ids == -100, pad_token_id)

        assert torch.all(shifted_input_ids >= 0).item(), "Verify that `shifted_input_ids` has only positive values"

        return shifted_input_ids

    def decode_custom(self,
                      labels=None,
                      decoder_input_ids=None,
                      decoder_attention_mask=None,
                      past_key_values=None,
                      encoder_hidden_states=None,
                      encoder_attention_mask=None,
                      head_mask=None,
                      use_cache=False,
                      output_attentions=False,
                      output_hidden_states=False,
                      return_dict=False, ):

        # Should always be the case?
        if labels is not None and decoder_input_ids is None:
            # get decoder inputs from shifting lm labels to the right
            decoder_input_ids = self._shift_right(labels)

        # If decoding with past key value states, only the last tokens
        # should be given as an input
        if past_key_values is not None:
            assert labels is None, "Decoder should not use cached key value states when training."
            if decoder_input_ids is not None:
                decoder_input_ids = decoder_input_ids[:, -1:]

        return self.decoder(input_ids=decoder_input_ids,
                            attention_mask=decoder_attention_mask,
                            past_key_values=past_key_values,
                            encoder_hidden_states=encoder_hidden_states,
                            encoder_attention_mask=encoder_attention_mask,
                            head_mask=head_mask,
                            use_cache=use_cache,
                            output_attentions=output_attentions,
                            output_hidden_states=output_hidden_states,
                            return_dict=return_dict, )



import os
import shutil
import subprocess

import torch
from torch.nn import functional as F
import transformers
from transformers import (CONFIG_NAME,
                          WEIGHTS_NAME,
                          GPT2LMHeadModel,
                          GPT2Tokenizer,
                          AutoTokenizer,
                          pipeline)


def load_pipeline(tokenizer, model_dir, pipeline_mode, redownload=False):
    model = GPT2LMHeadModel.from_pretrained(model_dir)
    device = 0 if torch.cuda.is_available() else -1
    print(f'Using device: {device}')
    P = pipeline(pipeline_mode, tokenizer = tokenizer, model = model, device = device)
    return P, model, device, model_dir






import numpy as np
from tqdm import tqdm
import time


class Annealer:
    def __init__(self, scorer, proposer, score_hyperparameters=dict(), proposer_hyperparameters=dict()):
        # Score output is {'total': num, 'constituent_1': num, ...'constituent_N':num}
        self.scorer = scorer

        self.proposer = proposer

        self.score_hyperparameters = score_hyperparameters
        self.proposer_hyperparameters = proposer_hyperparameters

    def propose_until_accepted(self, y, T):
        its = 1
        while True:
            proposal = self.proposer.propose(y, **self.proposer_hyperparameters)
            candidate_seq = proposal['output']
            candidate_score = self.scorer.score(candidate_seq, **self.score_hyperparameters)
            y_score = self.scorer.score(y, **self.score_hyperparameters)
            p_accept = min(1, (np.exp((candidate_score['total'] - y_score['total']) / T)))
            accepted = (np.random.uniform(0, 1) < p_accept)
            if accepted:
                return candidate_seq, its, candidate_score, y_score
            its += 1

    def anneal(self,
               y_init,
               max_search=1000,
               T_init=10,
               C=.1,
               eps=1 / 1e10,
               log=False,
               log_fn=lambda x: print(x),
               time_limit:float = 60):
        y = y_init
        y_hist = [y]
        t_hist = [T_init]
        it_hist = [0]
        candidate_score_hist = []
        y_score_hist = []

        start_time = time.time()
        for t in tqdm(range(max_search)):
            if log: log_fn(y)
            current_time = time.time() - start_time
            if (y == '') or current_time > time_limit:
                print(f'''Breaking Early.
                    compute time: {current_time}
                    y: {y}''')
                # Break early if converged to empty string or overrun time limit.
                break
            T = max(T_init - (C * t), eps)
            y, its, candidate_score, y_score = self.propose_until_accepted(y, T)
            y_hist.append(y)
            t_hist.append(T)
            it_hist.append(its)
            candidate_score_hist.append(candidate_score)
            y_score_hist.append(y_score)

        return {'y': y,
                'y_hist': y_hist,
                't_hist': t_hist,
                'it_hist': it_hist,
                'y_score_hist': y_score_hist,
                'candidate_score_hist': candidate_score_hist}



from abc import ABC, abstractmethod
from datetime import datetime

import numpy as np
from sklearn.metrics import accuracy_score, precision_recall_fscore_support
from torch.utils import data

from collators import DataCollatorForDocumentClassificationBATCH
from apteryx_datasets import PickleDatasetFromDisk
from transformers import (Trainer,
                          TrainingArguments)
from transformers import EarlyStoppingCallback

import dill as pickle
from pathlib import Path
import os
from sklearn.metrics import confusion_matrix


class AbstractMultiClassifier(ABC):
    def __init__(self, data_dir, model_name, block_size, training_args,
                 ds_limit, early_stopping_params, train_pct=.8, data_dir_is_ds=False, n_layers_to_train=0):
        self.model_name = model_name
        self.block_size = block_size
        self.n_layers_to_train = n_layers_to_train
        self.tokenizer = self.get_tokenizer_class().from_pretrained(self.model_name)
        self.collator = DataCollatorForDocumentClassificationBATCH()
        if not data_dir_is_ds:
            self.dataset = PickleDatasetFromDisk(data_dir=data_dir,
                                                 tokenizer=self.tokenizer,
                                                 block_size=self.block_size,
                                                 limit=ds_limit)
        else:
            self.dataset = data_dir

        self.config = LongformerConfig.from_pretrained(self.model_name,
                                                       vocab_size=self.tokenizer.vocab_size,
                                                       num_labels=self.dataset.num_classes)

        self.model = self.get_model_class().from_pretrained(self.model_name, config=self.config)

        if self.n_layers_to_train > 0:
            layer_acc = 0
            for layer in self.model.base_model.encoder.layer[:-self.n_layers_to_train]:
                print(f'Layer {layer_acc}: OFF')
                for p in layer.parameters():
                    p.requires_grad = False
                layer_acc += 1
            for layer in self.model.base_model.encoder.layer[-self.n_layers_to_train:]:
                print(f'Layer {layer_acc}: ON')
                for p in layer.parameters():
                    p.requires_grad = True
                layer_acc += 1
        else:
            # Turn all grads off except the final classification layer
            for i, p in enumerate(self.model.base_model.parameters()):
                # print(f'Layer {i}: OFF')
                p.requires_grad = False

        print(
            f'multi-classifier initialized; training Classification head and {self.n_layers_to_train} Attention Layers.')

        self.training_args = training_args
        self.early_stopping_params = early_stopping_params

        self.train_pct = train_pct

    def get_trainer(self):

        now = datetime.now()

        # Can use dicts and instantiate with ** instead of below:
        default_training_args = TrainingArguments(
            output_dir=f"{self.model_name.split('/')[-1]}-{now.day}-{now.month}-{now.year}",
            overwrite_output_dir=True,
            num_train_epochs=1,
            per_device_train_batch_size=32,
            save_steps=100,
            save_total_limit=2,
            logging_steps=20,
            dataloader_num_workers=15,
            warmup_steps=50,
            weight_decay=0.01,
            evaluation_strategy="steps",
            eval_steps=100,
            fp16=True,  # enable low-precision via AMP
            gradient_accumulation_steps=5
        )
        if self.training_args:
            # Update default training args with specified model training args.
            for k, v in self.training_args.items():
                print(f'Updating Trainer["{k}"] with: {v}.')
                setattr(default_training_args, k, v)

        self.training_args = default_training_args

        N = len(self.dataset)
        n_train = int(N * self.train_pct)
        n_val = N - n_train
        print(f'Training on {n_train} examples;')
        print(f'Validating on {n_val} examples.')
        train_ds, eval_ds = data.random_split(self.dataset, (n_train, n_val))

        callbacks = []
        if self.early_stopping_params:
            callbacks.append(EarlyStoppingCallback(**self.early_stopping_params))

        if len(callbacks) > 0:
            print('USING EARLY STOPPING.')
        else:
            print('NO EARLY STOPPING.')

        trainer = Trainer(model=self.model,
                          args=default_training_args,
                          data_collator=self.collator,
                          train_dataset=train_ds,
                          eval_dataset=eval_ds,
                          compute_metrics=self.compute_metrics,
                          callbacks=callbacks
                          )

        return trainer

    def train(self):
        trainer = self.get_trainer()
        trainer.train()

    @staticmethod
    def compute_metrics(pred):
        def save_confusion_matrix(mat):
            try:
                # Make a confusion matrix dir on desktop.
                outdir = Path(os.path.join(os.path.join(os.environ['HOME'], 'Desktop'), 'confusion_matrices'))
                if not os.path.exists(str(outdir)):
                    os.mkdir(str(outdir))

                now = datetime.now()
                f_string = f'{now.month}-{now.day}-{now.year}_{now.hour}hr-{now.minute}min-{now.second}sec'
                ext = '.pickle'
                fname = outdir / f_string
                fname = str(fname) + ext

                with open(fname, 'wb') as f:
                    pickle.dump(mat, f)
                    print(f'Saved confusion matrix to: {fname}')
            except:
                print('Encountered an error logging confusion matrix!')

        labels = pred.label_ids
        preds = pred.predictions.argmax(-1)
        precision, recall, f1s, _ = precision_recall_fscore_support(labels, preds, average='macro')
        acc = accuracy_score(labels, preds)

        try:
            confusion_matrix_to_log = confusion_matrix(labels, preds)
            save_confusion_matrix(confusion_matrix_to_log)




        except:
            print('compute_metrics broke trying to compute the confusion matrix! Skipping.')
            pass

        return {'accuracy': acc,
                'f1': f1,
                'precision': precision,
                'recall': recall}

    @abstractmethod
    def get_model_class(self):
        pass

    @abstractmethod
    def get_tokenizer_class(self):
        pass

    @abstractmethod
    def get_config_class(self):
        pass


# TODO: move this to another file?
from transformers import (LongformerConfig,
                          LongformerForSequenceClassification,
                          LongformerTokenizerFast)


class LongformerMultiClassifier(AbstractMultiClassifier):
    def __init__(self, data_dir, model_name='allenai/longformer-base-4096', block_size=4096,
                 training_args=None, ds_limit=np.inf, early_stopping_params=None, train_pct=0.8, data_dir_is_ds=False,
                 n_layers_to_train=0):
        super().__init__(data_dir, model_name, block_size, training_args, ds_limit,
                         early_stopping_params=early_stopping_params, train_pct=train_pct,
                         data_dir_is_ds=data_dir_is_ds, n_layers_to_train=n_layers_to_train)

    def get_model_class(self):
        return LongformerForSequenceClassification

    def get_tokenizer_class(self):
        return LongformerTokenizerFast

    def get_config_class(self):
        return LongformerConfig


# TODO: move this to another file?
from transformers import (RobertaConfig,
                          RobertaForSequenceClassification,
                          RobertaTokenizerFast)


class RobertaMultiClassifier(AbstractMultiClassifier):
    def __init__(self, data_dir, model_name='roberta-base', block_size=512, training_args=None, ds_limit=np.inf,
                 early_stopping_params=None, train_pct=0.8, data_dir_is_ds=False, n_layers_to_train=0):
        super().__init__(data_dir, model_name, block_size, training_args, ds_limit,
                         early_stopping_params=early_stopping_params, train_pct=train_pct,
                         data_dir_is_ds=data_dir_is_ds, n_layers_to_train=n_layers_to_train)

    def get_model_class(self):
        return RobertaForSequenceClassification

    def get_tokenizer_class(self):
        return RobertaTokenizerFast

    def get_config_class(self):
        return RobertaConfig



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Union
import os

import torch
from torch.utils import data
from transformers import (T5ForConditionalGeneration, T5Config, T5TokenizerFast)
from transformers import (Trainer,
                          TrainingArguments)
from transformers.modeling_outputs import (
    BaseModelOutput,
    BaseModelOutputWithPastAndCrossAttentions,
    Seq2SeqLMOutput,
    Seq2SeqModelOutput,
)

from ..collators import DataCollatorForAutoencodersBATCH
from .t5variants import T5EncoderAggDecoder


class AbstractTransformerAutoencoder(ABC):
    def __init__(self,
                 dataset,
                 model_name: str,
                 model_config_dict: dict,
                 training_args_dict: dict,
                 block_size: int,
                 tokenizer=None,
                 encoding_vector_size=512,
                 agg=True,
                 agg_mode='linear',
                 train_pct: float = .8, n_layers_to_train: Union[tuple, int] = 0):
        '''

        :param dataset: The PyTorch Dataset used for training.
        :param model_name: The huggingface model name (e.g. 't5-small')
        :param model_config_dict: A dict with model config args; will be passed to the config constructor.
        :param training_args_dict: A dict with training arguments; if nothing is provided, will use default training arguments.
        :param block_size: The input size of the model (in tokens).
        :param train_pct: The percentage of the dataset to use for training; the remainder will be used for eval.
        :param n_layers_to_train: The number of attention layers to train. If a tuple, specifies for encoder and decoder separately.
        '''
        self.model_name = model_name
        self.model_name_is_path = os.path.exists(self.model_name)
        if self.model_name_is_path:
            print(f'Loading model checkpoint from {self.model_name}')

        self.tokenizer = tokenizer if tokenizer else self.get_tokenizer_class().from_pretrained(self.model_name)
        self.collator = self.get_collator_class()()

        self.dataset = dataset
        self.block_size = self.dataset.block_size
        self.n_layers_to_train = n_layers_to_train

        if self.model_name_is_path:
            self.model = self.get_model_class().from_pretrained(self.model_name,
                                                                block_size=self.block_size,
                                                                encoding_vector_size=encoding_vector_size,
                                                                agg=agg,
                                                                agg_mode=agg_mode
                                                                )
            self.config = self.model.config
        else:
            self.config = self.get_config_class()(
                **model_config_dict) if model_config_dict else self.get_config_class()()
            self.config.decoder_start_token_id = self.tokenizer.pad_token_id

            # print(type(self.config))
            print(self.config)

            self.model = self.get_model_class()(config=self.config, block_size=self.block_size,
                                                encoding_vector_size=encoding_vector_size,
                                                agg=agg,
                                                agg_mode=agg_mode)
        self.encoder = self.model.encoder
        self.decoder = self.model.decoder

        self.n_enc_layers = len(self.model.encoder.block)
        self.n_dec_layers = len(self.model.decoder.block)
        self.total_attn_layers = self.n_enc_layers + self.n_dec_layers

        if isinstance(self.n_layers_to_train, tuple):
            if self.n_layers_to_train == (-1, -1):
                # In the case of (-1, -1), EVERYTHING requires grad.
                for p in self.model.parameters():
                    p.requires_grad = True
            else:
                n_enc_to_train, n_dec_to_train = self.n_layers_to_train
                assert n_enc_to_train <= self.n_enc_layers
                assert n_dec_to_train <= self.n_dec_layers

                self.toggle_layer_grad(self.model.encoder.block, n_enc_to_train, layer_acc=self.n_dec_layers)
                self.toggle_layer_grad(self.model.decoder.block, n_dec_to_train)

        elif isinstance(self.n_layers_to_train, int) and self.n_layers_to_train > 0:
            assert self.total_attn_layers >= self.n_layers_to_train, f"You must select a number of layers to train less than the total number of layers in the model. You selected {self.n_layers_to_train}; there are {self.total_attn_layers} attention layers available."
            n_dec_to_train = self.n_layers_to_train if self.n_layers_to_train < self.n_dec_layers else self.n_dec_layers
            n_enc_to_train = max(self.n_layers_to_train - self.n_dec_layers, 0)

            self.toggle_layer_grad(self.model.encoder.block, n_enc_to_train, layer_acc=self.n_dec_layers)
            self.toggle_layer_grad(self.model.decoder.block, n_dec_to_train)

        else:
            # Turn all grads off except the final classification layer
            for i, p in enumerate(self.model.parameters()):
                # print(f'Layer {i}: OFF')
                p.requires_grad = False

        print(f'Autoencoder initialized; training lm_head and {self.n_layers_to_train} Attention Layers.')

        self.training_args = training_args_dict

        self.train_pct = train_pct

    def toggle_layer_grad(self, module, n_to_train, layer_acc=0):
        n_layers = len(module)

        # Train all layers if n_to_train is -1.
        if n_to_train == -1:
            n_to_train = n_layers

        if n_to_train > 0:
            for layer in module[:-n_to_train]:
                print(f'Layer {layer_acc}: OFF')
                for p in layer.parameters():
                    p.requires_grad = False
                layer_acc += 1
            for layer in module[-n_to_train:]:
                print(f'Layer {layer_acc}: ON')
                for p in layer.parameters():
                    p.requires_grad = True
                layer_acc += 1

    def get_trainer(self):

        now = datetime.now()

        # Can use dicts and instantiate with ** instead of below:
        default_training_args = TrainingArguments(
            output_dir=f"{self.model_name.split('/')[-1]}-{now.day}-{now.month}-{now.year}",
            overwrite_output_dir=True,
            num_train_epochs=1,
            per_device_train_batch_size=32,
            save_steps=100,
            save_total_limit=2,
            logging_steps=20,
            dataloader_num_workers=15,
            warmup_steps=50,
            weight_decay=0.01,
            evaluation_strategy="steps",
            eval_steps=100,
            fp16=True,  # enable low-precision via AMP
            gradient_accumulation_steps=5
        )
        if self.training_args:
            # Update default training args with specified model training args.
            for k, v in self.training_args.items():
                print(f'Updating Trainer["{k}"] with: {v}.')
                setattr(default_training_args, k, v)

        self.training_args = default_training_args

        N = len(self.dataset)
        n_train = int(N * self.train_pct)
        n_val = N - n_train
        print(f'Training on {n_train} examples;')
        print(f'Validating on {n_val} examples.')
        train_ds, eval_ds = data.random_split(self.dataset, (n_train, n_val))

        trainer = Trainer(model=self.model,
                          args=default_training_args,
                          data_collator=self.collator,
                          train_dataset=train_ds,
                          eval_dataset=eval_ds)

        return trainer

    def train(self, checkpoint_dir=None):
        trainer = self.get_trainer()
        if checkpoint_dir:
            trainer.train(checkpoint_dir)
        else:
            trainer.train()

    @abstractmethod
    def get_config_class(self):
        pass

    @abstractmethod
    def get_model_class(self):
        pass

    @abstractmethod
    def get_tokenizer_class(self):
        pass

    @abstractmethod
    def get_collator_class(self):
        pass


class T5AutoEncoder(AbstractTransformerAutoencoder):
    def __init__(self, dataset, model_name='t5-base', block_size=1024,
                 encoding_vector_size=512,
                 tokenizer=None,
                 agg=True,
                 agg_mode='linear',
                 model_config_dict=None,
                 training_args_dict=None, train_pct=0.8,
                 n_layers_to_train=(-1, -1)):
        super().__init__(dataset=dataset, model_name=model_name, model_config_dict=model_config_dict,
                         training_args_dict=training_args_dict, block_size=block_size,
                         encoding_vector_size=encoding_vector_size,
                         tokenizer=tokenizer,
                         agg=agg,
                         agg_mode=agg_mode,
                         train_pct=train_pct,
                         n_layers_to_train=n_layers_to_train)

    def get_model_class(self):
        return T5EncoderAggDecoder

    def get_tokenizer_class(self):
        return T5TokenizerFast

    def get_config_class(self):
        return T5Config

    def get_collator_class(self):
        return DataCollatorForAutoencodersBATCH



import numpy as np
import spacy
import torch
import logging
import string

from transformers import RobertaTokenizerFast, RobertaForMaskedLM


class WordLevelProposer:
    def __init__(self,
                 tokenizer='roberta-base',
                 model='roberta-base',
                 spacy_model='en_core_web_sm',
                 device='cpu',
                 include_insert=True,
                 include_delete=False,
                 n_masks=1):
        self.device = device
        self.logger = logging.Logger('Proposer')
        self.nlp = spacy.load(spacy_model)
        self.tokenizer = RobertaTokenizerFast.from_pretrained(tokenizer)
        self.model = RobertaForMaskedLM.from_pretrained(model).to(self.device)
        self.ops = [('edit', self.edit), ('insert', self.insert) if include_insert else None,
                    ('delete', self.delete) if include_delete else None]
        self.ops = [i for i in self.ops if i]
        self.n_masks = n_masks

    def propose(self, s):
        opname, op = self.ops[np.random.randint(len(self.ops))]
        return {'op': opname,
                'output': op(s)}

    def edit(self, s, allow_same=False):
        masked, inputs, mask_idx, original_word = self.mask(s, mode='edit')
        new = inputs.input_ids.clone()
        out = self.model(**inputs)
        probs = out.logits.softmax(2)

        # Set probs of original vocab tokens to zero
        if not allow_same:
            masked_ids = self.tokenizer(original_word, add_special_tokens=False).input_ids
            probs[:, mask_idx, masked_ids] = 0

        edit = torch.multinomial(probs[:, mask_idx], num_samples=1, replacement=True).item()
        new[:, mask_idx] = edit

        return self.tokenizer.batch_decode(new, skip_special_tokens=True)[0]  # , mask_idx, original_word

    def insert(self, s):
        masked, inputs, mask_idx, original_word = self.mask(s, mode='insert')
        new = inputs.input_ids.clone()
        out = self.model(**inputs)

        edit = torch.multinomial(out.logits.softmax(2)[:, mask_idx], num_samples=1, replacement=True).item()
        new[:, mask_idx] = edit

        return self.tokenizer.batch_decode(new, skip_special_tokens=True)[0]  # , mask_idx, original_word

    def delete(self, s):
        '''
        This MOSTLY handles dangling whitespace after deletion. Revisit at some point - likely not a huge deal though.
        '''
        to_mask, start_idx, end_idx, pos, words = self.mask(s, mode='delete')

        left_char = s[start_idx - 1] if start_idx > 0 else ''
        right_char = s[end_idx] if end_idx < len(s) else ''

        n_whitespace_remaining = sum([left_char in string.whitespace, right_char in string.whitespace])

        if n_whitespace_remaining == 2:
            # Delete char to the right - it doesn't matter, both are whitespace
            return s[:start_idx] + s[end_idx + 1:]  # , to_mask

        elif n_whitespace_remaining == 1:
            # Don't delete anything
            return s[:start_idx] + s[end_idx:]  # , to_mask

        elif n_whitespace_remaining == 0:
            # Replace missing whitespace
            return s[:start_idx] + ' ' + s[end_idx:]  # , to_mask
        else:
            print('Should never reach here.')
            return s[:start_idx] + s[end_idx:]  # , to_mask

        # elif left_char in string.whitespace:
        #     #Right char is either punct or part of another word; remove in both directions
        #     return s[:start_idx - 1] + s[end_idx:]
        # elif right_char in string.whitespace:
        #     #Left char is likely punct; leave both
        #     return s[:start_idx] + s[end_idx:]

    def mask(self, s, mode='insert'):
        # encoded = self.tokenizer(s, return_tensors='pt')
        # original = encoded.input_ids
        # input_ids = original.clone()
        #
        # n_tok = input_ids.shape[-1]
        # #Randomly set a token to MASK
        # mask_idx = np.random.randint(1, n_tok - 1)
        # input_ids[:, mask_idx] = self.tokenizer.mask_token_id
        # encoded.input_ids = input_ids
        # encoded.update({'labels': original})
        # return encoded, mask_idx

        '''
        whole-word masking.
        '''
        words = [(token.text, token.idx, token.idx + len(token.text), token.pos_) for token in self.nlp(s)]
        word_idx = np.random.randint(len(words))
        to_mask, start_idx, end_idx, pos = words[word_idx]

        if mode == 'delete':
            return to_mask, start_idx, end_idx, pos, words

        if mode == 'insert':
            masked = s[:end_idx] + self.tokenizer.mask_token + s[end_idx:]

        elif mode == 'edit':
            masked = s[:start_idx] + self.tokenizer.mask_token + s[end_idx:]

        else:
            self.logger.error(f'Mask mode "{mode}" not valid')
            assert False

        inputs = self.tokenizer(masked, return_tensors="pt").to(self.device)

        mask_idx = (inputs.input_ids[0] == self.tokenizer.mask_token_id).nonzero().item()

        return masked, inputs, mask_idx, to_mask


class TokenLevelProposer:
    def __init__(self,
                 tokenizer='roberta-base',
                 model='roberta-base',
                 device='cpu',
                 include_insert=True,
                 include_delete=False):
        self.device = device
        self.logger = logging.Logger('Proposer')
        self.tokenizer = RobertaTokenizerFast.from_pretrained(tokenizer)
        self.model = RobertaForMaskedLM.from_pretrained(model).to(self.device)
        self.ops = ['edit',
                    'insert' if include_insert else None,
                    'delete' if include_delete else None]
        self.ops = [i for i in self.ops if i]

    def propose(self, s, n_masks=5):
        opname = self.ops[np.random.randint(len(self.ops))]
        return {'op': opname,
                'output': self._propose(s, mode=opname, n_masks=n_masks)}

    def _propose(self, s, mode='edit', n_masks=5):
        inputs, chosen_idxs = self.mask(s, mode=mode, n_masks=n_masks)
        assert inputs['input_ids'].shape[0] == 1, 'Did you accidentally pass in two strings?'

        if mode == 'delete':
            return self.tokenizer.batch_decode(inputs['input_ids'], skip_special_tokens=True)[0]
        else:
            out = self.model(**inputs)
            new = inputs['input_ids'].clone()

            probs = out.logits.softmax(2)
            mask_idxs = torch.nonzero(new == self.tokenizer.mask_token_id)

            #If there exist mask indexes:
            if mask_idxs.sum().item() != 0:
                mask_idxs = mask_idxs[:, 1]

                edits = torch.multinomial(probs[0][mask_idxs], num_samples=1, replacement=True)

                new[0][mask_idxs] = edits.flatten()
                return self.tokenizer.batch_decode(new, skip_special_tokens=True)[0]
            else:
                #Nothing was masked, so simply return s.
                return s

    def insert_mask_at(self, t1, idx):
        return torch.cat([t1[:, :idx], torch.Tensor([[self.tokenizer.mask_token_id]]), t1[:, idx:]], dim=1).long()

    def pop_at(self, t1, idxs_to_pop):
        all_idxs = np.arange(t1.shape[-1])
        return t1[:, [i.item() for i in all_idxs if i not in idxs_to_pop]]

    def mask(self, s, mode='insert', n_masks=5, max_sample_factor = .5):
        encoded = self.tokenizer(s, return_tensors='pt')
        original = encoded.input_ids
        input_ids = original.clone()

        # Choose tokens from 1 to end - 1 (avoid padding)
        # Make sure you don't try to sample too many tokens!
        # Never sample more than 1/2 of all the tokens in the inputs. This could be a hyp.


        n_to_chose = min(n_masks, int(input_ids.shape[-1] * max_sample_factor))


        options = np.arange(1, input_ids.shape[1] - 1)
        chosen_idxs = np.sort(np.random.choice(options, n_to_chose, replace=False)) if len(options.ravel()) >= n_to_chose else np.array([0])


        # Randomly set a token to MASK
        if mode == 'edit':
            try:
                input_ids[:, chosen_idxs] = self.tokenizer.mask_token_id
            except:
                print('Failed EDIT op.')

        elif mode == 'insert':
            try:
                # Work backwards (see [::-1] at end) so insertion doesn't cause offsets
                chosen_idxs = chosen_idxs[::-1]
                for idx in chosen_idxs:
                    input_ids = self.insert_mask_at(input_ids, idx)
            except:
                print('Failed INSERT op.')

        elif mode == 'delete':
            try:
                input_ids = self.pop_at(input_ids, chosen_idxs)
            except:
                print('Failed DELETE op.')

        to_return = {
            'input_ids': input_ids.to(device=self.device),
            'attention_mask': torch.ones(input_ids.shape).to(device=self.device)
        }

        return to_return, chosen_idxs



import json
from tqdm import tqdm

import torch
from transformers import TrainingArguments
from transformers import Trainer

from .smooth_gradient import SmoothGradient

from ..apteryx_datasets import BalancedDataset
from ..collators import DataCollatorForDocumentClassificationBATCH

# ----------------------------------------------------------------------------------------------------------------------------------


class Visualizer:
    def __init__(self, model, tokenizer, window_size, class_map:dict):
        self.model = model
        self.tokenizer = tokenizer
        self.window_size = window_size
        self.base_output_dir = './tmp'
        self.criterion = torch.nn.CrossEntropyLoss()
        self.collator = DataCollatorForDocumentClassificationBATCH()
        self.class_map = class_map

    def parse_relative_location_from_response(self, i, txt, offset):
        tok_probs = list(zip(i['tokens'], i['grad']))

        tmp = txt
        acc = 0
        start_end = []
        for tok, prob in tok_probs:
            tok = tok.replace("Ġ", "")
            tok = tok.replace("Ċ", "")
            start = tmp.index(tok)
            actual_start = acc + start
            end = start + len(tok)
            actual_end = acc + end
            tmp = tmp[end:]

            # print(tok, txt[actual_start:actual_end], start, end)
            start_end.append([tok, txt[actual_start:actual_end], prob, actual_start + offset, actual_end + offset])

            acc += end

        return start_end

    def visualize(self, txt, threshold=.7):
        '''
        :param txt: Document text. A string.
        :param threshold: Percentage of significance to report.
        :return: A colored HTML string.
        '''
        label = -1 #Only needed for underlying datashape. Should always be -1.

        ids = self.tokenizer(txt, add_special_tokens=False)['input_ids']
        chunked_data = [(self.tokenizer.decode(ids[i:i + self.window_size]), label)
                        for i in range(0, len(ids), self.window_size)]
        ds = BalancedDataset(self.tokenizer, chunked_data, block_size=self.window_size)
        training_args = TrainingArguments(
            output_dir=self.base_output_dir,  # TODO N: generate a better directory
            per_device_train_batch_size=2,
            local_rank=-1)
        trainer = Trainer(model=self.model,
                          args=training_args,
                          data_collator=self.collator,
                          train_dataset=ds)
        smooth_grad = SmoothGradient(self.model,
                                     self.criterion,
                                     self.tokenizer,
                                     show_progress=True,
                                     encoder="bert")
        instances = smooth_grad.saliency_interpret(trainer.get_train_dataloader())

        start_end_data = []
        chunk_len_acc = 0
        for idx, instance in enumerate(instances):
            try:
                chunk = chunked_data[idx][0]
                loc_data = self.parse_relative_location_from_response(instance, chunk, offset=chunk_len_acc)
                start_end_data.append(loc_data)
            except:
                continue
            chunk_len_acc += len(chunk)

        results = list()
        for i in instances:
            if i['prob'] > threshold:
                colored_string = smooth_grad.colorize(i, self.class_map)
                results.append(colored_string)
                # TODO 34: do something with this
        return results, instances, start_end_data



import numpy as np
import torch
from torch.nn.functional import softmax

from transformers import GPT2LMHeadModel, GPT2Config, GPT2TokenizerFast
from sentence_transformers import SentenceTransformer
from scipy.spatial.distance import cosine

import logging


class SemanticScorer:
    def __init__(self, target, model_name='stsb-roberta-large', device='cpu'):
        logger = logging.Logger('Semantic Scorer')

        self.model = SentenceTransformer(model_name).to(device)
        self.target = target
        #Cache t_emb to save compute.
        self.t_emb = self.model.encode(target)

    def __call__(self, sequence, target):
        return self.similarity(sequence, target)

    def similarity(self, sequence, target):
        s_emb = self.model.encode(sequence)
        return (1 / (cosine(s_emb, self.t_emb)) / np.pi)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['apteryx_transformers',
 'apteryx_transformers.models',
 'apteryx_transformers.parsers',
 'apteryx_transformers.simulated_annealing',
 'apteryx_transformers.simulated_annealing.heuristics',
 'apteryx_transformers.utils']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.4,<0.4.0',
 'fuzzysearch>=0.7.3,<0.8.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.1,<2.0.0',
 'python-Levenshtein>=0.12.2,<0.13.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.7.0,<2.0.0',
 'sentence-transformers>=2.0.0,<3.0.0',
 'setuptools>=57.4.0,<58.0.0',
 'spacy-transformers>=1.0.3,<2.0.0',
 'torch>=1.9.0,<2.0.0',
 'torchaudio>=0.9.0,<0.10.0',
 'torchvision>=0.10.0,<0.11.0',
 'tqdm>=4.61.2,<5.0.0']

setup_kwargs = {
    'name': 'apteryx-transformers',
    'version': '0.1.3',
    'description': '',
    'long_description': None,
    'author': 'mgbvox',
    'author_email': 'mgbvox@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import pandas as pd
import regex as re
import spacy
import string
import json
from copy import deepcopy

from apteryx_transformers.parsers.parser_utils import (remove_tables,
                                                       get_start_stop)
from apteryx_transformers.GLOBALS import BLOCKLIST, STRIPLIST

'''
^ : start at beginning of string
\W*? : match non-alphanumeric, but as little as possible
[0-9]+ : match numeric, at least one.
'''
NUM_PATTERN = re.compile('^\W*?([0-9]+[.A-z]*.*?[0-9]*)\W*?')

DEFAULT_CONFIG = {"window": 5, "remove_table_text": True, "severity": 0}


class NPParser:
    def __init__(self, spacy_model='en_core_web_lg', add_to_blocklist: list = [], config=DEFAULT_CONFIG):
        print('Initializing!')
        print(f'Loading spacy model: {spacy_model}')
        self.nlp = spacy.load(spacy_model)
        self.blocklist = BLOCKLIST + add_to_blocklist
        print(f'Blocklist: {self.blocklist}')
        self.config = config

    def clean_np(self, tokens):
        clean = [t.text_with_ws.upper() for t in tokens if not any([t.is_stop,
                                                                    # Allow hyphen exception
                                                                    (t.is_punct and not t.text == '-'),
                                                                    t.text_with_ws.lower() in self.blocklist])]
        if clean:
            # Join text and remove l/r whitespace, if any.
            clean_txt = ''.join(clean).strip()
            for s in STRIPLIST:
                clean_txt = clean_txt.replace(s.upper(), '')
            return clean_txt
        else:
            return None

    def set_config(self, new_config):
        self.config.update(new_config)
        return self.config

    def get_nps(self, s):
        start_stop = get_start_stop(self._get_nps(s, **self.config), s)

        return start_stop

    def _get_nps(self, s, window=5, remove_table_text=True, severity=0):
        '''
        Given an input string and a lookahead window, return a dataframe of detected noun phrases and their part numbers.
        '''
        s = str(s)
        if remove_table_text:
            s = remove_tables(s, severity=severity)

        doc = self.nlp(s)

        data = []
        for chunk in doc.noun_chunks:
            tok_end = chunk.end
            next_text = ''.join([t.text_with_ws for t in doc[tok_end:tok_end + window]])

            # Detect following number.
            num_match = re.match(NUM_PATTERN, next_text)
            if num_match:
                group = num_match.group(1)
                # Iteratively remove punctuation from the match, if found.
                while group[-1] in string.punctuation:
                    group = group[:-1]
                num_match = group

            data.append([chunk, num_match, next_text])

        df = pd.DataFrame.from_records(data)
        df.columns = ['np_raw', 'num', 'trailing']

        # Remove stopwords from np
        df['np_clean'] = df.np_raw.apply(self.clean_np)
        # Drop nps that were completely eliminated by the earlier step.
        df = df.dropna(subset=['np_clean', 'np_raw']).reset_index(drop=True)
        # Convert raw nps to text
        df['np_raw'] = df.np_raw.apply(lambda tokens: ''.join([t.text_with_ws for t in tokens]))

        df['in_blocklist'] = df.np_clean.apply(
            lambda noun_phrase: any([i.lower() in self.blocklist for i in re.split('\s', noun_phrase.strip())]))

        return df[~df.in_blocklist].dropna(subset=['in_blocklist']).reset_index(drop=True)

    def prep_report(self, df, group: str, cols: list):
        if len(df.dropna(subset=[group])) > 0:
            new_df = df.groupby(group).apply(lambda g: pd.Series([list(set(g[c].values)) for c in cols])).reset_index()
            new_df.columns = [group] + cols
            new_df['start'] = new_df.start.apply(lambda s: sorted(list(s)))
            new_df['end'] = new_df.end.apply(lambda s: sorted(list(s)))
            return new_df
        else:
            return None

    def report(self, s):
        nps = self.get_nps(s)
        np_groups = self.prep_report(nps, 'np_clean', ['num', 'np_raw', 'start', 'end'])
        num_groups = self.prep_report(nps, 'num', ['np_clean', 'np_raw', 'start', 'end'])

        to_return = {'main': nps.to_dict(orient='records'),
                     'nps': {'all': np_groups.to_dict(orient='records'),
                             'multiple': np_groups[np_groups.apply(lambda x: len(x.num) > 1, axis=1)].to_dict(
                                 orient='records')
                             } if isinstance(np_groups, pd.DataFrame) else None,
                     'nums': {'all': num_groups.to_dict(orient='records'),
                              'multiple': num_groups[num_groups.apply(lambda x: len(x.np_clean) > 1, axis=1)].to_dict(
                                  orient='records')
                              } if isinstance(num_groups, pd.DataFrame) else None
                     }

        return to_return

    '''
    Broken: handle None type
    '''

    def report_json(self, s):
        return self.report(s)
        # return serialize_report(self.report(s))


if __name__ == '__main__':
    p = NPParser(spacy_model='en_core_web_sm')
    txt = 'The present disclosure is concerned with a visual apparatus and a method for creation of artificial vision. In particular, the present disclosure provides an interface and method for controlling a visual prosthesis (i.e. device) implanted in an individual patient (i.e. subject) to create artificial vision. FIG. 1 shows a visual prosthesis apparatus'
    print(p.report_json(txt))



import gzip
import os
import numpy as np

import torch
from torch.utils.data.dataset import Dataset
from tqdm import tqdm
from pathlib import Path
import dill as pickle
import json


class AutoEncoder_JSONL_Dataset(Dataset):
    def __init__(self, data_dir, extract_op, tokenizer, block_size: int, limit=None, glob_pattern='*'):
        self.block_size = block_size
        self.tok = tokenizer
        self.DS_LIMIT = limit
        self.BLOCK_SIZE = block_size
        print('Ingesting data!')
        self.data = self.get_data(data_dir, extract_op, limit)

    def __len__(self):
        return self.data.labels.shape[0]

    def __getitem__(self, item):
        return {
            'input_ids': self.data.input_ids[item],
            'attention_mask': self.data.attention_mask[item],
            'labels': self.data.labels[item]
        }

    def get_data(self, data_dir, extract_op, limit):
        files = Path(data_dir).glob('*')

        jdata = []
        acc = 0
        for file_n, file in enumerate(files):
            print(f'File {file_n}: {file}')
            print(f'{acc} processed so far.')
            with open(file, 'r') as f:
                for line in tqdm(f.readlines()):
                    if acc < self.DS_LIMIT:
                        l = extract_op(json.loads(line))

                        enc = self.tok(l, return_tensors='pt')
                        ids = enc.input_ids

                        i_len = ids.shape[-1]
                        n_blocks = (i_len // self.BLOCK_SIZE)
                        trunc_len = n_blocks * self.BLOCK_SIZE

                        # Chunked ids
                        jdata.extend(self.tok.batch_decode((ids[:, :trunc_len].view(n_blocks, self.BLOCK_SIZE))))

                        acc += n_blocks
                    else:
                        tokenized = self.tok(jdata, padding='max_length',
                                        truncation=True,
                                        max_length=self.block_size,
                                        return_tensors='pt',
                                        add_special_tokens=False)

                        #Add labels for autoencoder!
                        tokenized.update({'labels': tokenized.input_ids})

                        return tokenized


class BalancedDataset(Dataset):
    def __init__(self, tokenizer, data, block_size: int, limit=None):
        self.block_size = block_size
        self.tok = tokenizer
        print('Ingesting data!')
        self.txt = [i[0] for i in tqdm(data[:limit])]
        self.labels = torch.tensor([i[1] for i in tqdm(data[:limit])])

    def __len__(self):
        return len(self.txt)

    def __getitem__(self, item):
        d = self.tok(self.txt[item],
                     padding='max_length',
                     truncation=True,
                     max_length=self.block_size,
                     return_tensors='pt',
                     add_special_tokens=False)
        d['labels'] = self.labels[item]
        return d


class PickleDataset(Dataset):
    def __init__(self, data_dir, tokenizer, block_size, limit=None):
        self.block_size = block_size
        self.tokenizer = tokenizer
        self.txt = []
        self.labels = []
        print(f'Ingesting Data from {data_dir}!')
        pickle_files = list(Path(data_dir).glob('*.pickle'))
        for f in tqdm(pickle_files[:limit]):
            with open(f, 'rb') as data:
                chunks = pickle.load(data)
                for t, l in chunks:
                    self.txt.append(t)
                    self.labels.append(l)
        self.labels = torch.Tensor(self.labels).long()

    @property
    def num_classes(self):
        return len(self.labels.unique())

    def __len__(self):
        return len(self.txt)

    def __getitem__(self, item):
        d = self.tokenizer(self.txt[item],
                           padding='max_length',
                           truncation=True,
                           max_length=self.block_size,
                           return_tensors='pt',
                           add_special_tokens=False)
        d['labels'] = self.labels[item]
        return d


# load from disk/pkl
# tokenize, get ids, chunk in block size, etc
# pad last chunk?
class ExpandedDataset(Dataset):
    # dataset: list of patent objects
    def __init__(self, tokenizer, dataset, class_labels, block_size: int, limit=None):
        self.block_size = block_size
        self.tokenizer = tokenizer
        self.class_map = {x: i for i, x in enumerate(class_labels)}
        print('Ingesting data!')
        self.labels = list()
        self.input_ids = list()
        for item in tqdm(dataset[:limit]):
            # 'publication_number', 'TC', 'pos', 'abs_text', 'desc_text', 'claims_text'
            item_txt = '\n'.join([item['abs_text'], item['desc_text'], item['claims_text']])
            tokenized_txt = self.tokenizer(item_txt, return_tensors='pt')  # input_ids, attention_mask
            input_ids = tokenized_txt['input_ids']
            q_tokens = input_ids.shape[1]
            # print('tokens in pat:', q_tokens)
            chunks = [input_ids[0, i:i + block_size]
                      for i in range(0, q_tokens, block_size)]
            for chunk in chunks:
                # print(chunk.shape)
                self.input_ids.append(chunk)
            # print('q chunks', len(chunks))
            label = self.class_map[int(item['TC'])]
            for i in range(len(chunks)):
                self.labels.append(label)

        self.labels = torch.tensor(self.labels)

    @property
    def num_classes(self):
        return len(self.labels.unique())

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, item):
        import torch.nn.functional as F
        input_ids = self.input_ids[item]
        mask = torch.ones(self.block_size)
        if len(input_ids) < self.block_size:
            delta = self.block_size - len(input_ids)
            input_ids = F.pad(input_ids,
                              [0, delta],
                              mode='constant',
                              value=self.tokenizer.pad_token_id)
            mask[-delta:] = 0
        return {'input_ids': input_ids,
                'attention_mask': mask,
                'labels': self.labels[item]}


class PickleDatasetFromDisk(Dataset):
    def __init__(self, data_dir, tokenizer, block_size, limit=np.inf):
        self.block_size = block_size
        self.tokenizer = tokenizer
        self.data_dir = data_dir

        print('Indexing dir...')
        self.index_map = dict()
        for i, p in tqdm(enumerate(Path(data_dir).glob('*/*.pickle'))):
            if i < limit:
                self.index_map[i] = p
            else:
                break

    @property
    def num_classes(self):
        return 8

    def __len__(self):
        return len(self.index_map)

    def __getitem__(self, item):
        # Map int item to file path
        chunk_file = self.index_map[item]
        with open(chunk_file, 'rb') as f:
            txt, label = pickle.load(f)

        label = torch.Tensor([label]).long()
        d = self.tokenizer(txt,
                           padding='max_length',
                           truncation=True,
                           max_length=self.block_size,
                           return_tensors='pt',
                           add_special_tokens=False)

        d['labels'] = label

        return d



class PickleDatasetByClass(Dataset):
    def __init__(self, index_path, data_dir, tokenizer, block_size, per_class_limit, is_autoencoder=False):
        assert 'by_class' in str(index_path), 'Not using a _by_class dataset!'
        self.block_size = block_size
        self.tokenizer = tokenizer
        self.index_path = index_path
        self.data_dir = data_dir
        self.is_autoencoder = is_autoencoder
        print('Loading index...')
        with open(self.index_path, 'rb') as f:
            self.class_to_patent_map = pickle.load(f)
        self.num_classes = len(self.class_to_patent_map)
        self.per_class_limit = per_class_limit

        print('Randomizing and truncating data...')
        for _class in tqdm(range(self.num_classes)):
            # Make sure no class has too few examples
            assert len(self.class_to_patent_map[
                           _class]) > self.per_class_limit, f"Make sure per_class_limit is less than {self.class_to_patent_map[_class]}."

            self.class_to_patent_map[_class] = np.random.choice(self.class_to_patent_map[_class],
                                                                size=self.per_class_limit, replace=False)


    def __len__(self):
        return self.num_classes * self.per_class_limit

    def __getitem__(self, item):
        _class = item // self.per_class_limit
        _class_idx = item % self.per_class_limit

        # Map int item to file path
        chunk_file = os.path.join(self.data_dir, self.class_to_patent_map[_class][_class_idx])
        with open(chunk_file, 'rb') as f:
            txt, label = pickle.load(f)
            assert int(_class) == int(label), f"Label {label} doesn't match Class {_class}!"

        label = torch.Tensor([label]).long()
        d = self.tokenizer(txt,
                           padding='max_length',
                           truncation=True,
                           max_length=self.block_size,
                           return_tensors='pt',
                           add_special_tokens=False)

        d['labels'] = d.input_ids if self.is_autoencoder else label

        return d


class GzippedPickleDatasetByClass(PickleDatasetByClass):
    def __init__(self, gz_index_path, data_dir, tokenizer, block_size, per_class_limit, absolute_limit=None):
        assert 'by_class' in str(gz_index_path), 'Not using a _by_class dataset!'
        self.block_size = block_size
        self.tokenizer = tokenizer
        self.gz_index_path = gz_index_path
        self.data_dir = data_dir
        print('Loading index...')
        with gzip.open(self.gz_index_path, 'rb') as f:
            self.class_to_patent_map = pickle.load(f)
        self.num_classes = len(self.class_to_patent_map)

        if absolute_limit:
            print('absolute_limit used - overriding per_class_limit if provided.')
            self.per_class_limit = int(absolute_limit / self.num_classes)
        else:
            self.per_class_limit = per_class_limit

        print('Randomizing and truncating data...')
        for _class in tqdm(range(self.num_classes)):
            # Make sure no class has too few examples
            assert len(self.class_to_patent_map[
                           _class]) > self.per_class_limit, f"Make sure per_class_limit is less than {self.class_to_patent_map[_class]}."

            self.class_to_patent_map[_class] = np.random.choice(self.class_to_patent_map[_class],
                                                                size=self.per_class_limit, replace=False)



import torch

from tqdm import tqdm

from .saliency_interpreter import SaliencyInterpreter


class SmoothGradient(SaliencyInterpreter):
    """
    Interprets the prediction using SmoothGrad (https://arxiv.org/abs/1706.03825)
    Registered as a `SaliencyInterpreter` with name "smooth-gradient".
    """
    def __init__(self,
                 model,
                 criterion,
                 tokenizer,
                 stdev=0.01,
                 num_samples=20,
                 show_progress=True,
                 **kwargs):
        super().__init__(model, criterion, tokenizer, show_progress, **kwargs)
        # Hyperparameters
        self.stdev = stdev
        self.num_samples = num_samples

    def saliency_interpret(self, test_dataloader):

        instances_with_grads = []
        iterator = tqdm(test_dataloader) if self.show_progress else test_dataloader

        for batch in iterator:

            # we will store there batch outputs such as gradients, probability, tokens
            # so as each of them are used in different places, for convenience we will create
            # it as attribute:
            self.batch_output = []
            self._smooth_grads(batch)
            batch_output = self.update_output()
            instances_with_grads.extend(batch_output)

        return instances_with_grads

    def _register_forward_hook(self, stdev: float):
        """
        Register a forward hook on the embedding layer which adds random noise to every embedding.
        Used for one term in the SmoothGrad sum.
        """

        def forward_hook(module, inputs, output):
            # Random noise = N(0, stdev * (max-min))
            scale = output.detach().max() - output.detach().min()
            noise = torch.randn(output.shape).to(output.device) * stdev * scale

            # Add the random noise
            output.add_(noise)

        # Register the hook
        embedding_layer = self.get_embeddings_layer()
        handle = embedding_layer.register_forward_hook(forward_hook)
        return handle

    def _smooth_grads(self, batch):
        total_gradients = None
        for _ in range(self.num_samples):
            handle = self._register_forward_hook(self.stdev)
            grads = self._get_gradients(batch)
            handle.remove()

            # Sum gradients
            if total_gradients is None:
                total_gradients = grads
            else:
                total_gradients = total_gradients + grads

        total_gradients /= self.num_samples

        self.batch_output.append(total_gradients)



import torch
from torch.nn.functional import softmax

import matplotlib
import matplotlib.pyplot as plt


class SaliencyInterpreter:
    def __init__(self,
                 model,
                 criterion,
                 tokenizer,
                 show_progress=True,
                 **kwargs):

        """
        :param model: nn.Module object - can be HuggingFace's model or custom one.
        :param criterion: torch criterion used to train your model.
        :param tokenizer: HuggingFace's tokenizer.
        :param show_progress: bool flag to show tqdm progress bar.
        :param kwargs:
            encoder: string indicates the HuggingFace's encoder, that has 'embeddings' attribute. Used
                if your model doesn't have 'get_input_embeddings' method to get access to encoder embeddings
        """

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model = model.to(self.device)
        self.model.eval()
        self.criterion = criterion
        self.tokenizer = tokenizer
        self.show_progress = show_progress
        self.kwargs = kwargs
        # to save outputs in saliency_interpret
        self.batch_output = None

    def _get_gradients(self, batch):
        # set requires_grad to true for all parameters, but save original values to
        # restore them later
        original_param_name_to_requires_grad_dict = {}
        for param_name, param in self.model.named_parameters():
            original_param_name_to_requires_grad_dict[param_name] = param.requires_grad
            param.requires_grad = True
        embedding_gradients = []
        hooks = self._register_embedding_gradient_hooks(embedding_gradients)

        loss = self.forward_step(batch)

        self.model.zero_grad()
        loss.backward()

        for hook in hooks:
            hook.remove()

        # restore the original requires_grad values of the parameters
        for param_name, param in self.model.named_parameters():
            param.requires_grad = original_param_name_to_requires_grad_dict[param_name]

        return embedding_gradients[0]

    def _register_embedding_gradient_hooks(self, embedding_gradients):
        """
        Registers a backward hook on the
        Used to save the gradients of the embeddings for use in get_gradients()
        When there are multiple inputs (e.g., a passage and question), the hook
        will be called multiple times. We append all the embeddings gradients
        to a list.
        """

        def hook_layers(module, grad_in, grad_out):
            embedding_gradients.append(grad_out[0])

        backward_hooks = []
        embedding_layer = self.get_embeddings_layer()
        backward_hooks.append(embedding_layer.register_backward_hook(hook_layers))
        return backward_hooks

    def get_embeddings_layer(self):
        if hasattr(self.model, "get_input_embeddings"):
            embedding_layer = self.model.get_input_embeddings()
        else:
            encoder_attribute = self.kwargs.get("encoder")
            assert encoder_attribute, "Your model doesn't have 'get_input_embeddings' method, thus you " \
                "have provide 'encoder' key argument while initializing SaliencyInterpreter object"
            embedding_layer = getattr(self.model, encoder_attribute).embeddings
        return embedding_layer

    def colorize(self, instance, class_map, skip_special_tokens=False):

        special_tokens = self.special_tokens

        word_cmap = matplotlib.cm.Wistia
        prob_cmap = matplotlib.cm.Wistia
        template = '<span class="barcode"; style="color: black; background-color: {}">{}</span>'
        colored_string = ''
        # Use a matplotlib normalizer in order to make clearer the difference between values
        normalized_and_mapped = matplotlib.cm.ScalarMappable(cmap=word_cmap).to_rgba(instance['grad'])
        for word, color in zip(instance['tokens'], normalized_and_mapped):
            if word in special_tokens and skip_special_tokens:
                continue
            # handle wordpieces
            word = word.replace("Ġ", " ")
            word = word.replace("Ċ", "\n")
            color = matplotlib.colors.rgb2hex(color[:3])
            colored_string += template.format(color, word)
        colored_string += template.format(0, "    Label: {} |".format(class_map[int(instance['label'])]))
        prob = instance['prob']
        color = matplotlib.colors.rgb2hex(prob_cmap(prob)[:3])
        colored_string += template.format(color, "{:.2f}%".format(instance['prob']*100)) + '|'
        return colored_string

    @property
    def special_tokens(self):
        """
        Some tokenizers don't have 'eos_token' and 'bos_token' attributes.
        So needed we some trick to get them.
        """
        if self.tokenizer.bos_token is None or self.tokenizer.eos_token is None:
            special_tokens = self.tokenizer.build_inputs_with_special_tokens([])
            special_tokens_ids = self.tokenizer.convert_ids_to_tokens(special_tokens)
            self.tokenizer.bos_token, self.tokenizer.eos_token = special_tokens_ids

        special_tokens = self.tokenizer.eos_token, self.tokenizer.bos_token
        return special_tokens

    def forward_step(self, batch):
        """
        If your model receive inputs in another way or you computing not
         like in this example simply override this method. It should return the batch loss
        :param batch: batch returned by dataloader
        :return: torch.Tensor: batch loss
        """
        input_ids = batch.get('input_ids').to(self.device)
        attention_mask = batch.get("attention_mask").to(self.device)
        outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)[0]

        label = torch.argmax(outputs, dim=1)
        batch_losses = self.criterion(outputs, label)
        loss = torch.mean(batch_losses)

        self.batch_output = [input_ids, outputs]

        return loss

    def update_output(self):
        """
        You can also override this method if you want to change the format
         of outputs. (e.g. store just gradients)
        :return: batch_output
        """

        input_ids, outputs, grads = self.batch_output

        probs = softmax(outputs, dim=-1)
        probs, labels = torch.max(probs, dim=-1)

        tokens = [
            self.tokenizer.convert_ids_to_tokens(input_ids_)
            for input_ids_ in input_ids
        ]

        embedding_grads = grads.sum(dim=2)
        # norm for each sequence
        norms = torch.norm(embedding_grads, dim=1, p=1)
        # normalizing
        for i, norm in enumerate(norms):
            embedding_grads[i] = torch.abs(embedding_grads[i]) / norm

        batch_output = []

        iterator = zip(tokens, probs, embedding_grads, labels)

        for example_tokens, example_prob, example_grad, example_label in iterator:
            example_dict = dict()
            # as we do it by batches we has a padding so we need to remove it
            example_tokens = [t for t in example_tokens if t != self.tokenizer.pad_token]
            example_dict['tokens'] = example_tokens
            example_dict['grad'] = example_grad.cpu().tolist()[:len(example_tokens)]
            example_dict['label'] = example_label.item()
            example_dict['prob'] = example_prob.item()
            batch_output.append(example_dict)
        return batch_output



from fuzzysearch import find_near_matches
from fuzzywuzzy import process
import string
import pandas as pd
import numpy as np
import json
from copy import deepcopy
import re


def fuzzy_extract(qs, ls, threshold):
    '''fuzzy matches 'qs' in 'ls' and returns list of
    tuples of (word,index)
    '''
    for word, _ in process.extractBests(qs, (ls,), score_cutoff=threshold):
        for match in find_near_matches(qs, word, max_l_dist=1):
            match = word[match.start:match.end]
            index = ls.find(match)
            yield (match, index)


def remove_tables(s, severity=0):
    n_ws = []
    n_alpha = []
    n_num = []
    n_total = []
    lines = []
    for line in s.split('\n'):
        n_alpha.append(sum([c in string.ascii_letters for c in line]))
        n_ws.append(sum([c in string.whitespace for c in line]))
        n_num.append(sum([c in string.digits for c in line]))
        n_total.append(len(line))
        lines.append(line)

    df = pd.DataFrame.from_records(zip(n_ws, n_alpha, n_num, n_total, lines))
    df.columns = ['ws', 'alpha', 'num', 'total', 'line']

    # anti-tableness heuristic: rewards dense alpha, penalize dense whitespace/numbers
    df['table_score'] = np.log(((df.alpha + 1) / (df.ws + df.num + 1)) * (df.total / len(df) + 1))

    return '\n'.join(df[df.table_score > severity].line.values)


def is_json(s):
    try:
        _ = json.loads(s)
        return True
    except:
        return False


def get_start_stop(df, txt):
    df[['start', 'end']] = None
    for noun_phrase in df.np_raw.unique():
        start_ends = [[m.start(), m.end()] for m in re.finditer(re.escape(noun_phrase), txt)]

        np_idxs = df[df.np_raw == noun_phrase].index.values
        for idx, (start, end) in zip(np_idxs, start_ends):
            df.loc[idx, 'start'] = start
            df.loc[idx, 'end'] = end
    return df
